/**************************************************************************   
 *									  *
 *    File "com_master.c"						  *
 *									  *
 *    Simple Matrix Multiplication for P4				  *
 * 									  *
 *    This master program takes a nxn matrix A  and nx1 matrix B 	  *
 *    from data file "data_matrix" to compute nx1 matrix C.		  *
 *    The matrix size should be a multiple of number of slaves.           *
 *									  *
 *    Slave program in   "com_slave.c" 					  *
 *    Process Group File "com.pg" 					  *
 **************************************************************************/


#include "p4.h"

#define         MAX     50    
#define         DATA    1      /* message type */

main(argc,argv)  
int argc;
char **argv;
{
  FILE *fp, *fopen(), *fclose();
  int A[MAX][MAX], B[MAX], result[MAX];
  int matrix_size, nslaves, type, size, id, to, from, my_id, i, j;
  int load, raw, send_raw;
  long start_time, end_time, elapsed_time;
  int *incoming;

  printf("Matrix Size : ");
  scanf("%d", &matrix_size);

  start_time = clock();
  p4_initenv(&argc,argv);
  p4_create_procgroup();
    
  nslaves = p4_num_total_ids() - 1;
  printf("Number of Slaves = %d\n",nslaves);
  my_id = p4_get_my_id();

  load = matrix_size / nslaves;

  /* get input data first */	
  fp = fopen("data_matrix", "r");
  if (fp==NULL) {
    printf("\n cannot find file ");
    exit(1);
  }

  /* get matrix A */
  printf("\n\n\n == MATRIX A == \n\n");
  for ( i = 0 ; i < matrix_size ; i++ ) {
    printf("\n"); 
    for ( j = 0 ; j < matrix_size ; j++ ){  
      fscanf(fp,"%d",&A[i][j]);
      printf("%7d ",A[i][j]);
    }
  }

  /* get matrix B */
  printf("\n\n\n == MATRIX B == \n\n");
  for ( j = 0 ; j < matrix_size ; j++ ) { 
    fscanf(fp,"%d",&B[j]);
    printf("%7d ",B[j]);
  }
  printf("\n"); 
  fclose(fp);       /* data ready */

  /* distribute data */

  /* send part of A matrix and B matrix */
  for ( to = 1 ; to <= nslaves ; to++ ) { 
    printf("Master is sending to processor %d\n", to);

    /* send the matrix size and load */
    size = sizeof(int);
    p4_sendr(DATA, to, &matrix_size,  size);
    p4_sendr(DATA, to, &load,         size);

    size = matrix_size * sizeof(int);
    for ( raw = 0 ; raw < load ; raw++ ) {
      send_raw = ( to - 1 ) * load + raw;
      p4_sendr(DATA, to, A[send_raw], size);
    }
    p4_sendr(DATA, to, B, size);  
  }

  /* receive the results from slave processes */
  type = DATA;
  size = load * sizeof(int); 
  incoming = NULL;
  for( from = 1 ; from <= nslaves ; from++ ) { 
    printf("Master is receiving from processor %d\n", from);
    p4_recv(&type, &from, &incoming, &size);
    for ( i = 0 ; i < load ; i++ ) {
      raw = ( from - 1 ) * load + i;
      result[raw] = incoming[i];
    }
    p4_msg_free(incoming);
  }

  p4_wait_for_end();

  end_time = clock();

  elapsed_time = end_time - start_time;

  printf("===================================\n");
  printf("  Total Time : %d\n", elapsed_time);
  printf("===================================\n");

  /* print the result */
  printf("\n\n\n == MATRIX C == \n\n");
  for( i = 0 ; i < matrix_size ; i++)
    printf("%7d ", result[i]);
  printf("\n");

  printf("\n\n Master Exiting \n");       

} /* end of main */






